m: int
n: int
from numpy import float32

import irspack.evaluator._core
import typing
import numpy
import scipy.sparse

_Shape = typing.Tuple[int, ...]

__all__ = ["EvaluatorCore", "Metrics"]

class EvaluatorCore:
    def __getstate__(self) -> tuple: ...
    def __init__(
        self,
        grount_truth: scipy.sparse.csr_matrix[numpy.float64],
        recommendable: typing.List[typing.List[int]],
    ) -> None: ...
    def __setstate__(self, arg0: tuple) -> None: ...
    def cache_X_as_set(self, arg0: int) -> None: ...
    def get_ground_truth(self) -> scipy.sparse.csr_matrix[numpy.float64]: ...
    def get_metrics_f32(
        self,
        score_array: numpy.ndarray[numpy.float32, _Shape[m, n]],
        cutoff: int,
        offset: int,
        n_threads: int,
        recall_with_cutoff: bool = False,
    ) -> Metrics: ...
    def get_metrics_f64(
        self,
        score_array: numpy.ndarray[numpy.float64, _Shape[m, n]],
        cutoff: int,
        offset: int,
        n_threads: int,
        recall_with_cutoff: bool = False,
    ) -> Metrics: ...
    pass

class Metrics:
    def __init__(self, arg0: int) -> None: ...
    def as_dict(self) -> typing.Dict[str, float]: ...
    def merge(self, arg0: Metrics) -> None: ...
    pass



import pickle
from inspect import isabstract

import numpy as np
import pytest
import scipy.sparse as sps

from irspack import BaseRecommender, Evaluator, get_recommender_class
from irspack.recommenders.base import RecommenderMeta

X_train = sps.csr_matrix(
    np.asfarray([[1, 1, 2, 0, 0], [0, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 0, 0]])
)

X_test = sps.csr_matrix(
    np.asfarray([[0, 0, 0, 1, 1], [1, 0, 0, 0, 0], [0, 0, 0, 0, 0], [1, 0, 0, 0, 0]])
)

rec_classes = list(RecommenderMeta.recommender_name_vs_recommender_class.keys())


@pytest.mark.parametrize("class_name", rec_classes)
def test_recs(class_name: str) -> None:
    """Test the learning of recommenders exit normally, and they are picklable.

    Args:
        class_name (str): The recommender class's name to be tested.
    """
    RecommenderClass = get_recommender_class(class_name)
    if isabstract(RecommenderClass):
        pytest.skip()
    rec = RecommenderClass.from_config(X_train, RecommenderClass.config_class())
    rec.learn()

    scores = rec.get_score(np.arange(X_train.shape[0]))
    eval = Evaluator(X_test, 0, 20)
    with pytest.raises(ValueError):
        eval.get_score(rec)
    metrics = eval.get_scores(rec, cutoffs=[X_train.shape[1]])
    assert np.all(np.isfinite(scores))
    assert np.all(~np.isnan(scores))
    for value in metrics.values():
        assert ~np.isnan(value)
        assert np.isfinite(value)
    with open("temp.pkl", "wb") as ofs:
        pickle.dump(rec, ofs)
    with open("temp.pkl", "rb") as ifs:
        rec_dumped: BaseRecommender = pickle.load(ifs)
    score_from_dumped = rec_dumped.get_score(np.arange(X_train.shape[0]))
    np.testing.assert_allclose(scores, score_from_dumped)



from irspack.recommenders.base import (
    BaseRecommender,
    BaseSimilarityRecommender,
    get_recommender_class,
)
from irspack.recommenders.base_earlystop import BaseRecommenderWithEarlyStopping
from irspack.recommenders.dense_slim import DenseSLIMConfig, DenseSLIMRecommender
from irspack.recommenders.edlae import EDLAERecommender
from irspack.recommenders.ials import IALSConfig, IALSRecommender
from irspack.recommenders.knn import (
    AsymmetricCosineKNNConfig,
    AsymmetricCosineKNNRecommender,
    CosineKNNConfig,
    CosineKNNRecommender,
    JaccardKNNConfig,
    JaccardKNNRecommender,
    TverskyIndexKNNConfig,
    TverskyIndexKNNRecommender,
)

from .p3 import P3alphaConfig, P3alphaRecommender
from .rp3 import RP3betaConfig, RP3betaRecommender
from .slim import SLIMConfig, SLIMRecommender
from .toppop import TopPopConfig, TopPopRecommender
from .user_knn import (
    AsymmetricCosineUserKNNConfig,
    AsymmetricCosineUserKNNRecommender,
    CosineUserKNNConfig,
    CosineUserKNNRecommender,
)

__all__ = [
    "BaseRecommender",
    "BaseSimilarityRecommender",
    "BaseRecommenderWithEarlyStopping",
    "TopPopConfig",
    "TopPopRecommender",
    "P3alphaConfig",
    "P3alphaRecommender",
    "RP3betaConfig",
    "RP3betaRecommender",
    "DenseSLIMConfig",
    "DenseSLIMRecommender",
    "SLIMConfig",
    "SLIMRecommender",
    "IALSConfig",
    "IALSRecommender",
    "CosineKNNConfig",
    "CosineKNNRecommender",
    "JaccardKNNConfig",
    "JaccardKNNRecommender",
    "TverskyIndexKNNConfig",
    "TverskyIndexKNNRecommender",
    "AsymmetricCosineKNNConfig",
    "AsymmetricCosineKNNRecommender",
    "CosineUserKNNConfig",
    "CosineUserKNNRecommender",
    "AsymmetricCosineUserKNNConfig",
    "AsymmetricCosineUserKNNRecommender",
    "get_recommender_class",
]

try:
    from irspack.recommenders.truncsvd import (
        TruncatedSVDConfig,
        TruncatedSVDRecommender,
    )

    __all__.extend(["TruncatedSVDRecommender", "TruncatedSVDConfig"])
except ImportError:  # pragma: no cover
    pass  # pragma: no cover

try:
    from irspack.recommenders.nmf import NMFConfig, NMFRecommender

    __all__.extend(["NMFRecommender", "NMFConfig"])
except ImportError:  # pragma: no cover
    pass  # pragma: no cover


try:
    from .multvae import MultVAEConfig, MultVAERecommender

    __all__.extend(["MultVAEConfig", "MultVAERecommender"])
except ImportError:  # pragma: no cover
    pass  # pragma: no cover

try:
    from .bpr import BPRFMConfig, BPRFMRecommender

    __all__.extend(["BPRFMConfig", "BPRFMRecommender"])
except ImportError:  # pragma: no cover
    pass  # pragma: no cover



import pickle
from collections import defaultdict
from io import BytesIO

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.evaluator import Evaluator, EvaluatorWithColdUser
from irspack.recommenders import P3alphaRecommender, TopPopRecommender
from irspack.recommenders.base import BaseRecommender
from irspack.split import rowwise_train_test_split


class MockRecommender(BaseRecommender, register_class=False):
    def __init__(self, X_all: sps.csr_matrix, scores: np.ndarray) -> None:
        super().__init__(X_all)
        self.scores = scores

    def get_score(self, user_indices: np.ndarray) -> np.ndarray:
        return self.scores[user_indices]

    def _learn(self) -> None:
        pass


@pytest.mark.parametrize(
    "U, I, dtype", [(10, 5, "float32"), (10, 30, "float64"), (3000, 5, "float32")]
)
def test_metrics(U: int, I: int, dtype: str) -> None:
    try:
        from sklearn.metrics import average_precision_score, ndcg_score
    except:
        pytest.skip()

    rns = np.random.RandomState(42)
    scores = rns.randn(U, I).astype(dtype)
    X_gt = (rns.rand(U, I) >= 0.7).astype(np.float64)
    eval = Evaluator(sps.csr_matrix(X_gt), offset=0, cutoff=I, n_threads=4)
    mock_rec = MockRecommender(sps.csr_matrix(X_gt.shape), scores)

    # float 16 not supported
    mock_rec_invalid = MockRecommender(
        sps.csr_matrix(X_gt.shape), scores.astype(np.float16)
    )
    with pytest.raises(ValueError):
        eval.get_score(mock_rec_invalid)
    my_score = eval.get_score(mock_rec)
    sklearn_metrics = defaultdict(list)
    for i in range(scores.shape[0]):
        if X_gt[i].sum() == 0:
            continue
        sklearn_metrics["map"].append(average_precision_score(X_gt[i], scores[i]))
        sklearn_metrics["ndcg"].append(ndcg_score(X_gt[i][None, :], scores[i][None, :]))

    for key in ["map", "ndcg"]:
        assert my_score[key] == pytest.approx(np.mean(sklearn_metrics[key]), abs=1e-8)

    with pytest.raises(ValueError):
        eval_emptymask = Evaluator(
            sps.csr_matrix(X_gt),
            offset=0,
            cutoff=I,
            n_threads=None,
            masked_interactions=sps.csr_matrix(
                (X_gt.shape[0] + 1, X_gt.shape[1])
            ),  # empty
        )

    X_gt = X_gt[(X_gt.sum(axis=1) > 0) & ((X_gt > 0).sum(axis=1) < I)]

    eval_emptymask = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=1,
        n_threads=None,
        mb_size=3,
        masked_interactions=sps.csr_matrix(X_gt.shape),  # empty
        recall_with_cutoff=True,
    )
    mock_rec = MockRecommender(sps.csr_matrix(X_gt), X_gt)
    perfect_score = eval_emptymask.get_score(mock_rec)
    assert perfect_score["recall"] == pytest.approx(1.0)
    eval_vicious = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=1,
        n_threads=1,
        masked_interactions=X_gt,
    )
    vicious_score = eval_vicious.get_score(mock_rec)
    assert vicious_score["recall"] == 0.0


@pytest.mark.parametrize("U, I, C", [(10, 5, 5), (10, 30, 29)])
def test_metrics_with_cutoff(U: int, I: int, C: int) -> None:
    try:
        from sklearn.metrics import ndcg_score
    except:
        pytest.skip()

    rns = np.random.RandomState(42)
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    eval = Evaluator(sps.csr_matrix(X_gt), offset=0, cutoff=C, n_threads=2)
    eval_finer_chunk = Evaluator(
        sps.csr_matrix(X_gt), offset=0, cutoff=C, n_threads=2, mb_size=1
    )
    # empty mask
    mock_rec = MockRecommender(sps.csr_matrix(X_gt.shape), scores)
    my_score = eval.get_score(mock_rec)
    my_score_finer = eval_finer_chunk.get_score(mock_rec)
    for key in my_score:
        assert my_score_finer[key] == pytest.approx(my_score[key])

    ndcg = 0.0
    valid_users = 0
    map = 0.0
    precision = 0.0
    recall = 0.0
    item_appearance_count = np.zeros((I,), dtype=np.float64)
    for i in range(U):
        nzs = set(X_gt[i].nonzero()[0])
        if len(nzs) == 0:
            continue
        valid_users += 1
        ndcg += ndcg_score(X_gt[[i]], scores[[i]], k=C)
        recommended = scores[i].argsort()[::-1][:C]
        recall_denom = min(C, len(nzs))
        ap = 0.0
        current_hit = 0
        for i, rec in enumerate(recommended):
            item_appearance_count[rec] += 1.0
            if rec in nzs:
                current_hit += 1
                ap += current_hit / float(i + 1)
        ap /= recall_denom
        map += ap
        recall += current_hit / recall_denom
        precision += current_hit / C
    entropy = (lambda p: -p.dot(np.log(p)))(
        item_appearance_count / item_appearance_count.sum()
    )
    item_appearance_sorted_normalized = (
        np.sort(item_appearance_count) / item_appearance_count.sum()
    )
    lorentz_curve = np.cumsum(item_appearance_sorted_normalized)

    gini_index = 0
    delta = 1 / I
    for i in range(I):
        f = 2 * (((i + 1) / I) - lorentz_curve[i])
        gini_index += delta * f

    assert my_score["ndcg"] == pytest.approx(ndcg / valid_users)
    assert my_score["map"] == pytest.approx(map / valid_users, abs=1e-8)
    assert my_score["precision"] == pytest.approx(precision / valid_users, abs=1e-8)
    assert my_score["recall"] == pytest.approx(recall / valid_users, abs=1e-8)
    assert my_score["entropy"] == pytest.approx(entropy)
    assert my_score["gini_index"] == pytest.approx(gini_index)


@pytest.mark.parametrize("U, I, U_test", [(10, 5, 3), (10, 30, 8)])
def test_metrics_colduser_mask(U: int, I: int, U_test: int) -> None:
    rns = np.random.RandomState(42)
    X_gt = (rns.rand(U, I) >= 0.5).astype(np.float64)
    X_gt = X_gt[(X_gt.sum(axis=1) > 0)]
    X_gt = sps.csr_matrix(X_gt)
    rec = TopPopRecommender(X_gt).learn()
    vicious_eval = EvaluatorWithColdUser(X_gt, X_gt, cutoff=1)
    vicious_metric = vicious_eval.get_score(rec)
    assert vicious_metric["hit"] == 0.0

    popularity = rec.get_score_cold_user(X_gt[:1, :]).ravel()
    most_pop_indices = np.where(popularity.max() == popularity)[0]

    X_gt_pop = np.zeros(X_gt.shape)
    X_gt_pop[:, most_pop_indices] = 1
    X_gt_pop = sps.csr_matrix(X_gt_pop)
    generous_eval = EvaluatorWithColdUser(
        X_gt_pop,
        X_gt_pop,
        cutoff=1,
        masked_interactions=sps.csr_matrix(X_gt.shape),
        recall_with_cutoff=True,
    )
    generous_metric = generous_eval.get_score(rec)
    assert generous_metric["recall"] == 1.0

    pickle_content = BytesIO()
    pickle.dump(generous_eval, pickle_content)

    pickle_content.seek(0)

    generous_eval_pickled = pickle.load(pickle_content)

    assert generous_eval_pickled.get_score(rec)["recall"] == 1.0


@pytest.mark.parametrize("U, I, U_test", [(10, 5, 3), (10, 30, 8)])
def test_metrics_ColdUser(U: int, I: int, U_test: int) -> None:
    rns = np.random.RandomState(42)
    uvec = rns.randn(U + U_test, 3)
    ivec = rns.randn(I, 3)
    true_score = uvec.dot(ivec.T)  # + rns.randn(U, I)
    X = sps.csr_matrix((true_score > 0).astype(np.float64))
    X_train = X[:U]
    X_val = X[U:]
    X_val_learn, X_val_target = rowwise_train_test_split(X_val, random_state=0)
    X_train_all = sps.vstack([X_train, X_val_learn])
    hot_evaluator = Evaluator(
        sps.csr_matrix(X_val_target), offset=U, cutoff=I // 2, n_threads=2
    )

    rec = P3alphaRecommender(X_train_all)
    rec.learn()
    hot_score = hot_evaluator.get_score(rec)
    with pytest.warns(UserWarning):
        cold_evaluator = EvaluatorWithColdUser(
            X_val_learn.tocsc(), X_val_target, cutoff=I // 2, mb_size=5
        )  # csc matrix input should raise warning about
        # memory ordering, as csc-csc matrix product will be csc,
        # hence col-major matrix when made dense.
        cold_score = cold_evaluator.get_score(rec)

    shuffle_index = np.arange(X_val_learn.shape[0])
    rns.shuffle(shuffle_index)
    cold_evaluator_shuffled = EvaluatorWithColdUser(
        X_val_learn[shuffle_index], X_val_target[shuffle_index], cutoff=I // 2
    )
    cold_score_shuffled = cold_evaluator_shuffled.get_score(rec)
    for key in cold_score:
        assert cold_score_shuffled[key] == pytest.approx(cold_score[key])

    for key in hot_score:
        assert hot_score[key] == pytest.approx(cold_score[key], abs=1e-8)


@pytest.mark.parametrize("U, I, C", [(10, 5, 5), (10, 30, 29)])
def test_recommender_check(U: int, I: int, C: int) -> None:
    rns = np.random.RandomState(42)
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    eval = Evaluator(sps.csr_matrix(X_gt), offset=0, cutoff=C, n_threads=2)
    mock_rec_too_few_users = MockRecommender(sps.csr_matrix((U - 1, I)), scores[1:])
    with pytest.raises(ValueError):
        eval.get_score(mock_rec_too_few_users)
    mock_rec_too_few_items = MockRecommender(sps.csr_matrix((U, I - 1)), scores[:, 1:])
    with pytest.raises(ValueError):
        eval.get_score(mock_rec_too_few_items)
    moc_rec_valid = MockRecommender(sps.csr_matrix((U, I)), scores)
    eval.get_score(moc_rec_valid)



import scipy.sparse as sps
from scipy.special import expit

from irspack.definitions import OptionalRandomState
from irspack.utils.random import convert_randomstate


def mf_example_data(
    n_users: int = 100,
    n_items: int = 100,
    n_components: int = 5,
    random_state: OptionalRandomState = None,
) -> sps.csr_matrix:
    rns = convert_randomstate(random_state)
    user_factors = rns.randn(n_users, n_components) / (n_components ** 0.5)
    item_factors = rns.randn(n_items, n_components) / (n_components ** 0.5)
    score = expit(user_factors.dot(item_factors.T))
    return sps.csr_matrix(rns.binomial(1, score))



import os
import time
import warnings
from tempfile import NamedTemporaryFile
from typing import Dict

import numpy as np
import pandas as pd
import pytest
from optuna import load_study
from optuna.storages import RDBStorage
from scipy import sparse as sps

from irspack import autopilot
from irspack.optimizers.autopilot import SameThreadBackend
from irspack.optimizers.base_optimizer import study_to_dataframe

from .mock_classes import (
    AutopilotMockEarlyStoppableRecommender,
    AutopilotMockEvaluator,
    AutopilotMockRecommender,
)

TIMESCALE = 1.5

X_small = sps.csr_matrix(
    (np.random.RandomState(42).rand(10, 1024) > 0.5).astype(np.float64)
)
X_answer = sps.csr_matrix(
    (np.random.RandomState(43).rand(*X_small.shape) > 0.5).astype(np.float64)
)


def test_autopilot_external_db() -> None:
    if os.environ.get("CIBUILDWHEEL") is not None:
        pytest.skip()
    evaluator = AutopilotMockEvaluator(X_answer)
    temp_db_file = NamedTemporaryFile()
    storage = RDBStorage(f"sqlite:///{temp_db_file.name}")
    with pytest.raises(ValueError):
        _ = autopilot(
            X_small,
            evaluator,
            memory_budget=1,
            n_trials=10,
            algorithms=["AutopilotMock"],
            timeout_singlestep=1,
            timeout_overall=5,
            storage=storage
            # task_resource_provider=SameThreadBackend,
        )
    _ = autopilot(
        X_small,
        evaluator,
        memory_budget=1,
        n_trials=10,
        algorithms=["AutopilotMock"],
        timeout_singlestep=1,
        timeout_overall=5,
        storage=storage,
        study_name="mock_study",
        task_resource_provider=SameThreadBackend,
    )
    study = load_study(storage=storage, study_name="mock_study")
    df = study_to_dataframe(study)
    # They are all executed on the same thread and no sigkill
    assert np.all(df["value"].values < 0)


def test_autopilot() -> None:
    if os.environ.get("CIBUILDWHEEL") is not None:
        pytest.skip()

    evaluator = AutopilotMockEvaluator(X_answer)
    recommender_class, best_param, trial_df = autopilot(
        X_small,
        evaluator,
        memory_budget=1,
        n_trials=10,
        algorithms=["AutopilotMock"],
        timeout_singlestep=2,
    )
    assert best_param["wait_time"] < 2.0
    assert trial_df.shape[0] == 10
    wait_times = trial_df["AutopilotMockOptimizer.wait_time"]
    assert np.all(trial_df.iloc[(wait_times.values > 2.0)]["ndcg@10"].isna())
    assert recommender_class is AutopilotMockRecommender


def test_autopilot_timeout() -> None:
    if os.environ.get("CIBUILDWHEEL") is not None:
        pytest.skip()

    evaluator = AutopilotMockEvaluator(X_answer)
    wait = 20
    with pytest.raises(RuntimeError):
        # no available algorithm
        autopilot(
            X_small,
            evaluator,
            memory_budget=1,
            n_trials=100,
            algorithms=["DenseSLIM"],
            timeout_overall=5,
            timeout_singlestep=1,
        )

    start = time.time()

    wait_times_given_by_callback: Dict[int, float] = {}

    def callback(trial_number: int, history_df: pd.DataFrame) -> None:
        wait_times_given_by_callback[trial_number] = history_df[
            "AutopilotMockOptimizer.wait_time"
        ].iloc[-1]

    recommender_class, best_param, trial_df = autopilot(
        X_small,
        evaluator,
        memory_budget=1,
        n_trials=100,
        algorithms=["AutopilotMock", "DenseSLIM"],
        timeout_overall=wait,
        timeout_singlestep=1,
        callback=callback,
    )
    end = time.time()
    assert wait < (end - start) + 1
    wait_times = trial_df["AutopilotMockOptimizer.wait_time"]
    assert np.all(trial_df.iloc[(wait_times.values > 1.0)]["ndcg@10"].isna())
    # dense slim should be skipped
    assert len({name for name in trial_df["optimizer_name"] if not pd.isna(name)}) == 1
    assert recommender_class is AutopilotMockRecommender
    recommender_class(X_small, **best_param).learn()
    for index, row in trial_df.iterrows():
        target_value = row["AutopilotMockOptimizer.wait_time"]
        test_value = wait_times_given_by_callback[index]
        if pd.isna(test_value):
            assert pd.isna(target_value)
        else:
            assert target_value == test_value


def test_autopilot_earlystop() -> None:
    if os.environ.get("CIBUILDWHEEL") is not None:
        pytest.skip()

    evaluator = AutopilotMockEvaluator(X_answer)
    recommender_class, best_param, trial_df = autopilot(
        X_small,
        evaluator,
        memory_budget=1,
        n_trials=20,
        algorithms=["AutopilotMockEarlyStoppable"],
        timeout_singlestep=2,
        random_seed=0,
    )
    assert trial_df.shape[0] == 20
    wait_times = trial_df["AutopilotMockEarlyStoppableOptimizer.wait_time"]
    no_result_location = np.where((5 * wait_times) > 2)[0]
    assert no_result_location.shape[0] > 0  # asserting this test is meaningful
    assert np.all(trial_df.iloc[no_result_location].value == 0)

    result_location = np.where(
        (5 * trial_df["AutopilotMockEarlyStoppableOptimizer.wait_time"])
        < 1.0  # They should reach at least epoch = 5
    )[0]
    if result_location.shape[0] > 0:  # asserting this test is meaningful
        assert np.all(trial_df.iloc[result_location].value < 0)
        assert recommender_class is AutopilotMockEarlyStoppableRecommender
    else:
        warnings.warn("No trials have reached epoch 5?")



from .movielens import (
    MovieLens1MDataManager,
    MovieLens20MDataManager,
    MovieLens100KDataManager,
)

__all__ = [
    "MovieLens100KDataManager",
    "MovieLens1MDataManager",
    "MovieLens20MDataManager",
]



import warnings
from collections import OrderedDict
from enum import Enum, auto
from typing import TYPE_CHECKING, Dict, List, Optional

import numpy as np
from scipy import sparse as sps

from irspack.definitions import DenseScoreArray, InteractionMatrix
from irspack.evaluator._core import EvaluatorCore, Metrics
from irspack.utils.threading import get_n_threads

if TYPE_CHECKING:
    from irspack.recommenders import base as base_recommender


class TargetMetric(Enum):
    ndcg = auto()
    recall = auto()
    hit = auto()
    map = auto()
    precision = auto()


METRIC_NAMES = [
    "hit",
    "recall",
    "ndcg",
    "map",
    "precision",
    "gini_index",
    "entropy",
    "appeared_item",
]


class Evaluator:
    r"""Evaluates recommenders' performance against validation set.

    Args:
        ground_truth (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            The ground-truth.
        offset (int):
            Where the validation target user block begins.
            Often the validation set is defined for a subset of users.
            When offset is not 0, we assume that the users with validation
            ground truth corresponds to X_train[offset:] where X_train
            is the matrix feeded into the recommender class. Defaults to 0.
        cutoff (int, optional):
            Controls the default number of recommendation.
            When the evaluator is used for parameter tuning, this cutoff value will be used.
            Defaults to 10.
        target_metric (str, optional):
            Specifies the target metric when this evaluator is used for
            parameter tuning. Defaults to "ndcg".
        recommendable_items (Optional[List[int]], optional):
            Global recommendable items. Defaults to None.
            If this parameter is not None, evaluator will be concentrating on
            the recommender's score output for these recommendable_items,
            and compute the ranking performance within this subset.
        per_user_recommendable_items (Optional[List[List[int]]], optional):
            Similar to `recommendable_items`, but this time the recommendable items can vary among users. Defaults to None.
        masked_interactions (Optional[Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]], optional):
            If set, this matrix masks the score output of recommender model where it is non-zero.
            If none, the mask will be the training matrix itself owned by the recommender.

        n_threads (int, optional):
            Specifies the Number of threads to sort scores and compute the evaluation metrics.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if the variable is not set, it will be set to ``os.cpu_count()``. Defaults to None.

        recall_with_cutoff (bool, optional):
            This affects the definition of recall.
            If ``True``, for each user, recall will be computed as

            .. math ::

                \frac{N_{\text{hit}}}{\min(\text{cutoff}, N_{\text{ground truth}})}

            If ``False``, this will be

            .. math ::

                \frac{N_{\text{hit}}}{N_{\text{ground truth}}}


        mb_size (int, optional):
            The rows of chunked user score. Defaults to 1024.
    """

    n_users: int
    n_items: int
    masked_interactions: Optional[sps.csr_matrix]

    def __init__(
        self,
        ground_truth: InteractionMatrix,
        offset: int = 0,
        cutoff: int = 10,
        target_metric: str = "ndcg",
        recommendable_items: Optional[List[int]] = None,
        per_user_recommendable_items: Optional[List[List[int]]] = None,
        masked_interactions: Optional[InteractionMatrix] = None,
        n_threads: Optional[int] = None,
        recall_with_cutoff: bool = False,
        mb_size: int = 128,
    ) -> None:

        ground_truth = ground_truth.tocsr().astype(np.float64)
        ground_truth.sort_indices()
        if recommendable_items is None:
            if per_user_recommendable_items is None:
                recommendable_items_arg: List[List[int]] = []
            else:
                recommendable_items_arg = per_user_recommendable_items
        else:
            recommendable_items_arg = [recommendable_items]

        self.core = EvaluatorCore(ground_truth, recommendable_items_arg)
        self.offset = offset
        self.n_users = ground_truth.shape[0]
        self.n_items = ground_truth.shape[1]
        self.target_metric = TargetMetric[target_metric]
        self.cutoff = cutoff
        self.target_metric_name = f"{self.target_metric.name}@{self.cutoff}"
        self.n_threads = get_n_threads(n_threads)
        self.mb_size = mb_size
        if masked_interactions is None:
            self.masked_interactions = None
        else:
            if masked_interactions.shape != ground_truth.shape:
                raise ValueError(
                    "ground_truth and masked_interactions have different shapes. "
                )
            self.masked_interactions = sps.csr_matrix(masked_interactions)

        self.recall_with_cutoff = recall_with_cutoff

    def _get_metrics(
        self, scores: DenseScoreArray, cutoff: int, ground_truth_begin: int
    ) -> Metrics:
        if scores.dtype == np.float64:
            return self.core.get_metrics_f64(
                scores,
                cutoff,
                ground_truth_begin,
                self.n_threads,
                self.recall_with_cutoff,
            )
        elif scores.dtype == np.float32:
            return self.core.get_metrics_f32(
                scores,
                cutoff,
                ground_truth_begin,
                self.n_threads,
                self.recall_with_cutoff,
            )
        else:
            raise ValueError("score must be either float32 or float64.")

    def get_target_score(self, model: "base_recommender.BaseRecommender") -> float:
        r"""Compute the optimization target score (self.target_metric) with the cutoff being ``self.cutoff``.

        Args:
            model: The evaluated model.

        Returns:
            The metric value.
        """
        return self.get_score(model)[self.target_metric.name]

    def get_score(self, model: "base_recommender.BaseRecommender") -> Dict[str, float]:
        r"""Compute the score with the cutoff being ``self.cutoff``.

        Args:
            model : The evaluated recommender.

        Returns:
            metric values.
        """
        return self._get_scores_as_list(model, [self.cutoff])[0]

    def get_scores(
        self, model: "base_recommender.BaseRecommender", cutoffs: List[int]
    ) -> Dict[str, float]:
        r"""Compute the score with the specified cutoffs.

        Args:
            model : The evaluated recommender.
            cutoffs : for each value in cutoff, the class computes
                the metric values.

        Returns:
            The Resulting metric values. This time, the result
            will look like ``{"ndcg@20": 0.35, "map@20": 0.2, ...}``.
        """

        result: Dict[str, float] = OrderedDict()
        scores = self._get_scores_as_list(model, cutoffs)
        for cutoff, score in zip(cutoffs, scores):
            for metric_name in METRIC_NAMES:
                result[f"{metric_name}@{cutoff}"] = score[metric_name]
        return result

    def _get_scores_as_list(
        self, model: "base_recommender.BaseRecommender", cutoffs: List[int]
    ) -> List[Dict[str, float]]:
        if self.offset + self.n_users > model.n_users:
            raise ValueError("evaluator offset + n_users exceeds the model's n_users.")
        if self.n_items != model.n_items:
            raise ValueError("The model and evaluator assume different n_items.")
        n_items = self.n_items
        metrics: List[Metrics] = []
        for c in cutoffs:
            metrics.append(Metrics(n_items))

        block_start = self.offset
        n_validated = self.n_users
        block_end = block_start + n_validated
        mb_size = self.mb_size

        for chunk_start in range(block_start, block_end, mb_size):
            chunk_end = min(chunk_start + mb_size, block_end)
            try:
                # try faster method
                scores = model.get_score_block(chunk_start, chunk_end)
            except NotImplementedError:
                # block-by-block
                scores = model.get_score(np.arange(chunk_start, chunk_end))

            if self.masked_interactions is None:
                mask = model.X_train_all[chunk_start:chunk_end]
            else:
                mask = self.masked_interactions[
                    chunk_start - self.offset : chunk_end - self.offset
                ]
            scores[mask.nonzero()] = -np.inf
            for i, c in enumerate(cutoffs):
                chunked_metric = self._get_metrics(
                    scores,
                    c,
                    chunk_start - self.offset,
                )
                metrics[i].merge(chunked_metric)

        return [item.as_dict() for item in metrics]


class EvaluatorWithColdUser(Evaluator):
    r"""Evaluates recommenders' performance against cold (unseen) users.

    Args:
        input_interaction (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            The cold-users' known interaction with the items.
        ground_truth (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            The held-out ground-truth.
        offset (int): Where the validation target user block begins.
            Often the validation set is defined for a subset of users.
            When offset is not 0, we assume that the users with validation
            ground truth corresponds to X_train[offset:] where X_train
            is the matrix feeded into the recommender class.
        cutoff (int, optional):
            Controls the number of recommendation.
            Defaults to 10.
        target_metric (str, optional):
            Optimization target metric.
            Defaults to "ndcg".
        recommendable_items (Optional[List[int]], optional):
            Global recommendable items. Defaults to None.
            If this parameter is not None, evaluator will be concentrating on
            the recommender's score output for these recommendable_items,
            and compute the ranking performance within this subset.
        per_user_recommendable_items (Optional[List[List[int]]], optional):
            Similar to `recommendable_items`, but this time the recommendable items can vary among users. Defaults to None.
        masked_interactions (Optional[Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]], optional):
            If set, this matrix masks the score output of recommender model where it is non-zero.
            If none, the mask will be the training matrix (``input_interaction``) it self.
        n_threads (int, optional):
            Specifies the Number of threads to sort scores and compute the evaluation metrics.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if the variable is not set, it will be set to ``os.cpu_count()``. Defaults to None.
        recall_with_cutoff (bool, optional):
            This affects the definition of recall.
            If ``True``, for each user, recall will be evaluated by

            .. math ::

                \frac{N_{\text{hit}}}{\min( \text{cutoff}, N_{\text{ground truth}} )}

            If ``False``, this will be

            .. math ::

                \frac{N_{\text{hit}}}{N_{\text{ground truth}}}

        mb_size (int, optional):
            The rows of chunked user score. Defaults to 1024.
    """

    def __init__(
        self,
        input_interaction: InteractionMatrix,
        ground_truth: InteractionMatrix,
        cutoff: int = 10,
        target_metric: str = "ndcg",
        recommendable_items: Optional[List[int]] = None,
        per_user_recommendable_items: Optional[List[List[int]]] = None,
        masked_interactions: Optional[InteractionMatrix] = None,
        n_threads: Optional[int] = None,
        recall_with_cutoff: bool = False,
        mb_size: int = 1024,
    ):

        super().__init__(
            ground_truth,
            offset=0,
            cutoff=cutoff,
            target_metric=target_metric,
            recommendable_items=recommendable_items,
            per_user_recommendable_items=per_user_recommendable_items,
            masked_interactions=masked_interactions,
            n_threads=n_threads,
            recall_with_cutoff=recall_with_cutoff,
            mb_size=mb_size,
        )
        self.input_interaction = input_interaction

    def _get_scores_as_list(
        self,
        model: "base_recommender.BaseRecommender",
        cutoffs: List[int],
    ) -> List[Dict[str, float]]:

        n_items = model.n_items
        metrics: List[Metrics] = []
        for c in cutoffs:
            metrics.append(Metrics(n_items))

        block_start = self.offset
        n_validated = self.n_users
        block_end = block_start + n_validated
        mb_size = self.mb_size

        for chunk_start in range(block_start, block_end, mb_size):
            chunk_end = min(chunk_start + mb_size, block_end)
            scores = model.get_score_cold_user(
                self.input_interaction[chunk_start:chunk_end]
            )
            if self.masked_interactions is None:
                mask = self.input_interaction[chunk_start:chunk_end]
            else:
                mask = self.masked_interactions[chunk_start:chunk_end]
            scores[mask.nonzero()] = -np.inf

            if not scores.flags.c_contiguous:
                warnings.warn(
                    "Found col-major(fortran-style) score values.\n"
                    "Transforming it to row-major score matrix."
                )
                scores = np.ascontiguousarray(scores, dtype=np.float64)

            for i, c in enumerate(cutoffs):
                chunked_metric = self._get_metrics(scores, c, chunk_start)
                metrics[i].merge(chunked_metric)

        return [item.as_dict() for item in metrics]



import pickle
import time
from typing import IO, Any, Dict, List, Optional

import numpy as np
import scipy.sparse as sps
from scipy.special import expit

from irspack import (
    BaseOptimizer,
    BaseRecommender,
    DenseScoreArray,
    InteractionMatrix,
    TargetMetric,
    UserIndexArray,
)
from irspack.evaluator.evaluator import Evaluator
from irspack.optimizers.base_optimizer import BaseOptimizerWithEarlyStopping
from irspack.parameter_tuning import IntegerSuggestion, LogUniformSuggestion, Suggestion
from irspack.recommenders.base_earlystop import (
    BaseRecommenderWithEarlyStopping,
    TrainerBase,
)


class AutopilotMockRecommender(BaseRecommender, register_class=False):
    def __init__(self, X: InteractionMatrix, wait_time: float = 1.0):
        super().__init__(X)
        self.wait_time = wait_time

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        raise NotImplementedError("This is a mock")

    def _learn(self) -> None:
        time.sleep(self.wait_time)


class AutopilotMockOptimizer(BaseOptimizer):
    recommender_class = AutopilotMockRecommender
    default_tune_range: List[Suggestion] = [
        LogUniformSuggestion("wait_time", 1e-2, 4.0)
    ]

    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_in_mb: int
    ) -> List[Suggestion]:
        return []


class AutopilotMockEarlyStoppableTrainer(TrainerBase):
    def __init__(self, wait_time: float) -> None:
        self.epoch: int = 0
        self.wait_time = wait_time

    def load_state(self, ifs: IO) -> None:
        self.epoch = pickle.load(ifs)["epoch"]

    def save_state(self, ofs: IO) -> None:
        pickle.dump(dict(epoch=self.epoch), ofs)

    def run_epoch(self) -> None:
        time.sleep(self.wait_time)
        self.epoch += 1


class AutopilotMockEarlyStoppableRecommender(
    BaseRecommenderWithEarlyStopping, register_class=False
):
    trainer_class = AutopilotMockEarlyStoppableTrainer

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        wait_time: float,
        target_epoch: int,
        **kwargs: Any
    ):
        self.wait_time = wait_time
        self.target_epoch = target_epoch
        super().__init__(X_train_all)

    def _create_trainer(self) -> TrainerBase:
        return AutopilotMockEarlyStoppableTrainer(self.wait_time)

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        raise NotImplementedError("This is a mock")


class AutopilotMockEarlyStoppableOptimizer(BaseOptimizerWithEarlyStopping):
    recommender_class = AutopilotMockEarlyStoppableRecommender
    default_tune_range = [
        IntegerSuggestion("target_epoch", 5, 30, 5),
        LogUniformSuggestion("wait_time", 1e-2, 1e2),
    ]


class AutopilotMockEvaluator(Evaluator):
    def __init__(self, X: sps.csr_matrix) -> None:
        super().__init__(X)
        self.target_metric = TargetMetric.ndcg
        self.cutoff = 10

    def get_score(self, model: BaseRecommender) -> Dict[str, float]:
        if isinstance(model, AutopilotMockRecommender):
            return {self.target_metric.name: expit(model.wait_time)}
        elif isinstance(model, AutopilotMockEarlyStoppableRecommender):
            assert model.trainer is not None
            assert isinstance(model.trainer, AutopilotMockEarlyStoppableTrainer)
            score = np.exp(-abs(model.trainer.epoch - model.target_epoch))
            return {self.target_metric.name: score}

        assert False, "should not happern"



import numpy as np

from irspack.definitions import OptionalRandomState


def convert_randomstate(arg: OptionalRandomState) -> np.random.RandomState:
    if arg is None or isinstance(arg, int):
        return np.random.RandomState(arg)
    elif isinstance(arg, np.random.RandomState):
        return arg
    else:
        raise ValueError(f"{arg} cannot be interpreted as a random state.")



import gc

import numpy as np
from scipy import linalg, sparse

from ..definitions import InteractionMatrix
from .base import BaseSimilarityRecommender, RecommenderConfig


class EDLAEConfig(RecommenderConfig):
    reg: float = 1.0
    dropout_p: float = 0.1


class EDLAERecommender(BaseSimilarityRecommender):
    """Implementation of EDLAE (Emphasized Denoising Linear Autoencoder).

    See:
        - `Autoencoders that don't overfit towards the Identity (NeurIPS 2020)`
          by Harald Steck (NetFlix)

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.

        reg (float, optional):
            The L2 constant regularization parameter. Defaults to 1.0.

        dropout_p (float, optional):
            Probability of dropout. Defaults to 0.1
    """

    config_class = EDLAEConfig

    def __init__(
        self, X_train_all: InteractionMatrix, reg: float = 1.0, dropout_p: float = 0.1
    ):

        super(EDLAERecommender, self).__init__(X_train_all)
        self.reg = reg
        self.dropout_p = dropout_p

    def _learn(self) -> None:
        X_train_all_f32 = self.X_train_all.astype(np.float32)

        P = X_train_all_f32.T.dot(X_train_all_f32)
        P_dense: np.ndarray = P.todense()
        del P
        q = 1 - self.dropout_p
        lamb = self.dropout_p / q * np.diag(P_dense) + self.reg
        P_dense[np.arange(self.n_items), np.arange(self.n_items)] += lamb
        gc.collect()
        P_dense = linalg.inv(P_dense, overwrite_a=True)

        gc.collect()
        diag_P_inv = 1 / np.diag(P_dense)
        P_dense *= -diag_P_inv[np.newaxis, :]
        range_ = np.arange(self.n_items)
        P_dense[range_, range_] = 0
        self.W_ = P_dense



from typing import Any, List, Optional, Tuple

import numpy as np
import pandas as pd
import scipy.sparse as sps

from irspack.definitions import InteractionMatrix, OptionalRandomState
from irspack.utils.id_mapping import IDMappedRecommender
from irspack.utils.random import convert_randomstate
from irspack.utils.threading import get_n_threads

from ._util_cpp import (
    okapi_BM_25_weight,
    remove_diagonal,
    rowwise_train_test_split_by_fixed_n,
    rowwise_train_test_split_by_ratio,
    sparse_mm_threaded,
    tf_idf_weight,
)


def l1_normalize_row(X: sps.csc_matrix) -> sps.csc_matrix:
    result: sps.csc_matrix = X.astype(np.float64)
    result.sort_indices()
    l1_norms: np.ndarray = result.sum(axis=1).A1
    result.data /= l1_norms[result.indices]
    return result


def rowwise_train_test_split(
    X: InteractionMatrix,
    test_ratio: float = 0.5,
    n_test: Optional[int] = None,
    ceil_n_test: bool = False,
    random_state: OptionalRandomState = None,
) -> Tuple[InteractionMatrix, InteractionMatrix]:
    """Splits the non-zero elements of a sparse matrix into two (train & test interactions).
    For each row, the ratio of non-zero elements that become the test interaction
    is (approximately) constant.

    Args:
        X:
            The source sparse matrix.
        test_ratio:
            The ratio of test interactions for each row.
            That is, for each row, if it contains ``NNZ``-nonzero elements,
            the number of elements entering into the test interaction
            will be ``math.floor(test_ratio * NNZ)``.
            Defaults to 0.5.
        random_state:
            The random state. Defaults to `None`.

    Returns:
        A tuple of train & test interactions, which sum back to the original matrix.
    """
    rns = convert_randomstate(random_state)
    random_seed = rns.randint(0, np.iinfo(np.int32).max, dtype=np.int32)
    original_dtype = X.dtype
    X_double = X.astype(np.float64)
    if n_test is None:
        X_train_double, X_test_double = rowwise_train_test_split_by_ratio(
            X_double, random_seed, test_ratio, ceil_n_test
        )
    else:
        X_train_double, X_test_double = rowwise_train_test_split_by_fixed_n(
            X_double, random_seed, n_test
        )
    return (
        X_train_double.astype(original_dtype),
        X_test_double.astype(original_dtype),
    )


def df_to_sparse(
    df: pd.DataFrame,
    user_colname: str,
    item_colname: str,
    user_ids: Optional[List[Any]] = None,
    item_ids: Optional[List[Any]] = None,
    rating_colname: Optional[str] = None,
) -> Tuple[sps.csr_matrix, np.ndarray, np.ndarray]:
    r"""Convert pandas dataframe into sparse matrix.

    Args:
        df:
            The dataframe to be converted into a sparse matrix.
        user_colname:
            The column name for users.
        item_colname:
            The column name for items.
        user_ids:
            If not `None`, the resulting matrix's rows correspond exactly to this list.
        item_ids:
            If not `None`, the resulting matrix's columns correspond exactly to this list.
        rating_colname:
            If not `None`, the non-zero elements of the resulting matrix will correspond to the values of this column.
    Raises:
        RuntimeError:
            If `user_ids` is not `None` and `df[user_colname]` contains values not in `user_ids`.
        RuntimeError:
            If `item_ids` is not `None` and `df[item_colname]` contains values not in `item_ids`.

    Returns:
        - The resulting sparse matrix.
        - user ids corresponding to the rows in the matrix.
        - item ids corresponding to the columns in the matrix.
    """
    user_codes = pd.Categorical(df[user_colname], categories=user_ids)
    item_codes = pd.Categorical(df[item_colname], categories=item_ids)
    if np.any(user_codes.codes < 0):
        raise RuntimeError("Found unknown user id.")
    if np.any(item_codes.codes < 0):
        raise RuntimeError("Found unknown item id.")
    row = user_codes.codes
    unique_user_ids = user_codes.categories
    col = item_codes.codes
    unique_item_ids = item_codes.categories
    if rating_colname is None:
        data = np.ones(df.shape[0])
    else:
        data = np.asfarray(df[rating_colname].values)
    return (
        sps.csr_matrix(
            (data, (row, col)), shape=(len(unique_user_ids), len(unique_item_ids))
        ),
        unique_user_ids,
        unique_item_ids,
    )


__all__ = [
    "l1_normalize_row",
    "rowwise_train_test_split",
    "sparse_mm_threaded",
    "okapi_BM_25_weight",
    "tf_idf_weight",
    "remove_diagonal",
    "get_n_threads",
    "IDMappedRecommender",
    "convert_randomstate",
]



import re
import time
import warnings
from abc import ABC, abstractmethod
from logging import Logger
from multiprocessing import Pipe as mp_pipe
from multiprocessing import Process
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Type
from uuid import uuid1

import numpy as np
import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.storages import RDBStorage
from optuna.trial import TrialState

from irspack.definitions import InteractionMatrix
from irspack.evaluator import Evaluator
from irspack.optimizers.base_optimizer import (
    LowMemoryError,
    get_optimizer_class,
    study_to_dataframe,
)
from irspack.parameter_tuning.parameter_range import Suggestion, is_valid_param_name
from irspack.recommenders import BaseRecommender
from irspack.utils.default_logger import get_default_logger

DEFAULT_SEARCHNAMES = ["RP3beta", "IALS", "DenseSLIM", "AsymmetricCosineKNN", "SLIM"]

_sort_intermediate: Callable[[Tuple[int, float]], float] = lambda _: _[1]


def search_one(
    pipe: Any,
    X: InteractionMatrix,
    evaluator: Evaluator,
    optimizer_names: List[str],
    suggest_overwrites: Dict[str, List[Suggestion]],
    db_url: str,
    study_name: str,
    random_seed: int,
    logger: Logger,
) -> None:
    study = optuna.load_study(
        storage=db_url,
        study_name=study_name,
        sampler=TPESampler(seed=random_seed),
    )

    def _obj(trial: optuna.Trial) -> float:
        pipe.send(trial.number)
        optimizer_name = trial.suggest_categorical("optimizer_name", optimizer_names)
        assert isinstance(optimizer_name, str)

        optimizer = get_optimizer_class(optimizer_name)(
            X,
            evaluator,
            suggest_overwrite=suggest_overwrites[optimizer_name],
            logger=logger,
        )
        return optimizer.objective_function(optimizer_name + ".")(trial)

    study.optimize(_obj, n_trials=1)


class TaskBackend(ABC):
    @abstractmethod
    def __init__(
        self,
        X: InteractionMatrix,
        evaluator: Evaluator,
        optimizer_names: List[str],
        suggest_overwrites: Dict[str, List[Suggestion]],
        db_url: str,
        study_name: str,
        random_seed: int,
        logger: Logger,
    ):
        raise NotImplementedError()

    @property
    def exit_code(self) -> Optional[int]:
        return self._exit_code()

    @abstractmethod
    def receive_trial_number(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def _exit_code(self) -> Optional[int]:
        raise NotImplementedError()

    @abstractmethod
    def start(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def join(self, timeout: Optional[int]) -> None:
        raise NotImplementedError()

    @abstractmethod
    def terminate(self) -> None:
        raise NotImplementedError()  # pragma: no cover


class MultiProcessingBackend(TaskBackend):
    def __init__(
        self,
        X: InteractionMatrix,
        evaluator: Evaluator,
        optimizer_names: List[str],
        suggest_overwrites: Dict[str, List[Suggestion]],
        db_url: str,
        study_name: str,
        random_seed: int,
        logger: Logger,
    ):
        self.pipe_parent, pipe_child = mp_pipe()
        self._p = Process(
            target=search_one,
            args=(
                pipe_child,
                X,
                evaluator,
                optimizer_names,
                suggest_overwrites,
                db_url,
                study_name,
                random_seed,
                logger,
            ),
        )

    def _exit_code(self) -> Optional[int]:
        return self._p.exitcode

    def receive_trial_number(self) -> int:
        result: int = self.pipe_parent.recv()
        return result

    def start(self) -> None:
        self._p.start()

    def join(self, timeout: Optional[int]) -> None:
        self._p.join(timeout=timeout)

    def terminate(self) -> None:
        self._p.terminate()


class SameThreadBackend(TaskBackend):
    def __init__(
        self,
        X: InteractionMatrix,
        evaluator: Evaluator,
        optimizer_names: List[str],
        suggest_overwrites: Dict[str, List[Suggestion]],
        db_url: str,
        study_name: str,
        random_seed: int,
        logger: Logger,
    ):
        self._args = (
            X,
            evaluator,
            optimizer_names,
            suggest_overwrites,
            db_url,
            study_name,
            random_seed,
            logger,
        )
        self._trial_number: Optional[int] = None

    def send(self, i: int) -> None:
        self._trial_number = i

    def start(self) -> None:
        search_one(self, *self._args)

    def receive_trial_number(self) -> int:
        assert self._trial_number is not None
        return self._trial_number

    def join(self, timeout: Optional[int]) -> None:
        warnings.warn("Single step timeout will be ignored for ThreadingBackend.")
        pass

    def terminate(self) -> None:
        # can't do that
        pass

    def _exit_code(self) -> Optional[int]:
        # must always normally exit
        return 0


def autopilot(
    X: InteractionMatrix,
    evaluator: Evaluator,
    n_trials: int = 20,
    memory_budget: int = 4000,  # 4GB
    timeout_overall: Optional[int] = None,
    timeout_singlestep: Optional[int] = None,
    algorithms: List[str] = DEFAULT_SEARCHNAMES,
    random_seed: Optional[int] = None,
    logger: Optional[Logger] = None,
    callback: Optional[Callable[[int, pd.DataFrame], None]] = None,
    storage: Optional[RDBStorage] = None,
    study_name: Optional[str] = None,
    task_resource_provider: Type[TaskBackend] = MultiProcessingBackend,
) -> Tuple[Type[BaseRecommender], Dict[str, Any], pd.DataFrame]:

    r"""Given an interaction matrix and an evaluator, search for the best algorithm and its parameters
    (roughly) within the time & space constraints. You can specify how each search step will be executed.

    Args:
        X:
            Input interaction matrix.
        evaluator:
            Evaluator to measure the performance of the recommenders.
        n_trials: The maximal number of trials. Defaults to 20.
        memory_budget:
            Optimizers will try search parameters so that memory usage (in megabyte) will not exceed this values.
            An algorithm will not be searched if it inevitably violates this bound.
            Note that this value is quite rough one and will not be respected strictly.
        timeout_overall:
            If set, the total execution time of the trials will not exceed this value (roughly).
        timeout_singlestep:
            If set, a single trial (recommender and a set of its parameter) will not run for more than the value (in seconds).
            Such a trial is considered to have produced  a score value of 0,
            and optuna will avoid suggesting such values (if everything works fine).
            Defaults to `None`.
        algorithms:
            A list of algorithm names to be tried.
            Defaults to `["RP3beta", "IALS", "DenseSLIM", "AsymmetricCosineKNN", "SLIM"]`.
        random_seed:
            The random seed that controls the suggestion behavior.
            Defaults to `None`.
        logger:
            The logger to be used. If `None`, irspack's default logger will be used.
            Defaults to None.
        callback:
            If not `None`, called at the end of every single trial with the following arguments

                1. The current trial's number.
                2. A `pd.DataFrame` that holds history of trial execution.

            Defaults to `None`.
        storage:
            An instance of `optuna.storages.RDBStorage`. Defaults to `None`.
        study_name:
            If `storage` argument is given, you have to pass study_name
            argument.
        task_resource_provider:
            Specifies how each search step is executed. Defaults to `MultiProceesingBackend`.
    Raises:
        ValueError:
            If `storage` is given but `study_name` is not specified.
        RuntimeError:
            If no recommender algorithms are available within given memory budget.
        RuntimeError:
            If no trials have been completed within given timeout.


    Returns:

        * The best algorithm's recommender class.
        * The best parameters.
        * The dataframe containing the history of trials.

    """
    if storage is not None and study_name is None:
        raise ValueError('"study_name" must be specified if "storage" is given.')
    RNS = np.random.RandomState(random_seed)
    suggest_overwrites: Dict[str, List[Suggestion]] = {}
    optimizer_names: List[str] = []
    for rec_name in algorithms:
        optimizer_class_name = rec_name + "Optimizer"
        optimizer_class = get_optimizer_class(optimizer_class_name)
        try:
            suggest_overwrites[
                optimizer_class_name
            ] = optimizer_class.tune_range_given_memory_budget(X, memory_budget)
            optimizer_names.append(optimizer_class_name)
        except LowMemoryError:
            continue

    if not optimizer_names:
        raise RuntimeError("No available algorithm with given memory.")

    if logger is None:
        logger = get_default_logger()

    logger.info("Trying the following algorithms: %s", optimizer_names)

    optional_db_path = Path(f".autopilot-{uuid1()}.db")
    storage_: RDBStorage
    if storage is None:
        storage_ = RDBStorage(
            url=f"sqlite:///{optional_db_path.name}",
        )
    else:
        storage_ = storage

    if study_name is None:
        study_name_ = f"autopilot-{uuid1()}"
    else:
        study_name_ = study_name
    start = time.time()
    study = optuna.create_study(
        storage=storage_, study_name=study_name_, load_if_exists=True
    )
    study_id = storage_.get_study_id_from_name(study_name_)

    for _ in range(n_trials):

        task_start = time.time()
        elapsed_at_start = task_start - start

        timeout_for_this_process: Optional[int] = None
        if timeout_overall is None:
            timeout_for_this_process = timeout_singlestep
        else:
            remaining_time = int(timeout_overall - elapsed_at_start)
            if remaining_time <= 0:
                break

            if timeout_singlestep is not None:
                timeout_for_this_process = min(remaining_time, timeout_singlestep)
        task = task_resource_provider(
            X,
            evaluator,
            optimizer_names,
            suggest_overwrites,
            storage_.url,
            study_name_,
            RNS.randint(0, np.iinfo(np.int32).max, dtype=np.int32),
            logger,
        )

        task.start()
        trial_number = task.receive_trial_number()
        task.join(timeout=timeout_for_this_process)

        if task.exit_code is None:
            task.terminate()
            try:
                logger.info(f"Trial {trial_number} timeout.")
                storage_.read_trials_from_remote_storage(study_id)
                trial_id = storage_.get_trial_id_from_study_id_trial_number(
                    study_id, trial_number
                )
                trial_this = storage_.get_trial(trial_id)
                intermediate_values = sorted(
                    list(trial_this.intermediate_values.items()),
                    key=_sort_intermediate,
                )

                if intermediate_values:
                    # Though terminated, it resulted in some values.
                    # Regard it as a COMPLETE trial.
                    storage_.set_trial_values(
                        trial_id,
                        [intermediate_values[0][1]],
                    )
                    storage_.set_trial_user_attr(
                        trial_id, "max_epoch", intermediate_values[0][0] + 1
                    )
                else:
                    # Penalize such a time-consuming trial
                    storage_.set_trial_values(trial_id, [0.0])
                storage_.set_trial_state(trial_id, TrialState.COMPLETE)
            except RuntimeError:  # pragma: no cover
                pass  # pragma: no cover

        if callback is not None:
            callback(trial_number, study_to_dataframe(study))

        now = time.time()
        elapsed = now - start
        if timeout_overall is not None:
            if elapsed > timeout_overall:
                break
    best_params_with_prefix = dict(
        **study.best_trial.params,
        **{
            key: val
            for key, val in study.best_trial.user_attrs.items()
            if is_valid_param_name(key)
        },
    )
    best_params = {
        re.sub(r"^([^\.]*\.)", "", key): value
        for key, value in best_params_with_prefix.items()
    }
    optimizer_name: str = best_params.pop("optimizer_name")
    result_df = study_to_dataframe(study)

    if storage is None:
        optional_db_path.unlink()
    recommender_class = get_optimizer_class(optimizer_name).recommender_class

    return (recommender_class, best_params, result_df)



import enum
from abc import abstractmethod
from typing import Optional, Union

from ..definitions import InteractionMatrix
from ..utils import get_n_threads, okapi_BM_25_weight, remove_diagonal, tf_idf_weight
from ._knn import (
    AsymmetricSimilarityComputer,
    CosineSimilarityComputer,
    JaccardSimilarityComputer,
    TverskyIndexComputer,
)
from .base import BaseSimilarityRecommender, RecommenderConfig


class FeatureWeightingScheme(str, enum.Enum):
    NONE = "NONE"
    TF_IDF = "TF_IDF"
    BM_25 = "BM_25"


class BaseKNNConfig(RecommenderConfig):
    shrinkage: float = 0.0
    top_k: int = 100
    n_threads: Optional[int] = None


class BaseKNNRecommender(BaseSimilarityRecommender):
    def __init__(
        self,
        X_train_all: InteractionMatrix,
        shrinkage: float = 0.0,
        top_k: int = 100,
        n_threads: Optional[int] = None,
        feature_weighting: str = "NONE",
        bm25_k1: float = 1.2,
        bm25_b: float = 0.75,
    ):
        super().__init__(X_train_all)
        self.shrinkage = shrinkage
        self.top_k = top_k
        self.feature_weighting = FeatureWeightingScheme(feature_weighting)
        self.bm25_k1 = bm25_k1
        self.bm25_b = bm25_b
        self.n_threads = get_n_threads(n_threads)

    @abstractmethod
    def _create_computer(
        self, X: InteractionMatrix
    ) -> Union[
        CosineSimilarityComputer,
        AsymmetricSimilarityComputer,
        JaccardSimilarityComputer,
        TverskyIndexComputer,
    ]:
        raise NotImplementedError("")

    def _learn(self) -> None:
        if self.feature_weighting == FeatureWeightingScheme.NONE:
            X_weighted = self.X_train_all
        elif self.feature_weighting == FeatureWeightingScheme.TF_IDF:
            X_weighted = tf_idf_weight(self.X_train_all)
        elif self.feature_weighting == FeatureWeightingScheme.BM_25:
            X_weighted = okapi_BM_25_weight(self.X_train_all, self.bm25_k1, self.bm25_b)
        else:
            raise RuntimeError("Unknown weighting scheme.")

        computer = self._create_computer(X_weighted.T)
        self.W_ = remove_diagonal(
            computer.compute_similarity(self.X_train_all.T, self.top_k)
        ).tocsc()


class BaseCosineKNNConfig(BaseKNNConfig):
    feature_weighting: str = "NONE"
    bm25_k1: float = 1.2
    bm25_b: float = 0.75


class CosineKNNConfig(BaseCosineKNNConfig):
    normalize: bool = False


class CosineKNNRecommender(BaseKNNRecommender):
    r"""K-nearest neighbor recommender system based on cosine similarity. That is, the similarity matrix ``W`` is given by (column-wise top-k restricted)

    .. math::

        \mathrm{W}_{i,j} = \begin{cases}
            \frac{\sum_{u} X_{ui} X_{uj}}{||X_{*i}||_2 ||X_{*j}||_2 + \mathrm{shrinkage}} & (\text{if normalize = True}) \\
            \sum_{u} X_{ui} X_{uj} & (\text{if normalize = False})
        \end{cases}

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        shrinkage (float, optional):
            The shrinkage parameter for regularization. Defaults to 0.0.
        normalize (bool, optional):
            Whether to normalize the similarity. Defaults to False.
        top_k (int, optional):
            Specifies the maximal number of allowed neighbors. Defaults to 100.
        feature_weighting (str, optional):
            Specifies how to weight the feature. Must be one of:

                - "NONE" : no feature weighting
                - "TF_IDF" : TF-IDF weighting
                - "BM_25" : `Okapi BM-25 weighting <https://en.wikipedia.org/wiki/Okapi_BM25>`_

            Defaults to "NONE".
        bm25_k1 (float, optional):
            The k1 parameter for BM25. Ignored if ``feature_weighting`` is not "BM_25". Defaults to 1.2.
        bm25_b (float, optional):
            The b parameter for BM25. Ignored if ``feature_weighting`` is not "BM_25". Defaults to 0.75.
        n_threads (Optional[int], optional): Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if the variable is not set, it will be set to ``os.cpu_count()``. Defaults to None.

    """

    config_class = CosineKNNConfig

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        shrinkage: float = 0.0,
        normalize: bool = False,
        top_k: int = 100,
        feature_weighting: str = "NONE",
        bm25_k1: float = 1.2,
        bm25_b: float = 0.75,
        n_threads: Optional[int] = None,
    ) -> None:

        super().__init__(
            X_train_all,
            shrinkage,
            top_k,
            n_threads,
            feature_weighting=feature_weighting,
            bm25_k1=bm25_k1,
            bm25_b=bm25_b,
        )
        self.normalize = normalize

    def _create_computer(self, X: InteractionMatrix) -> CosineSimilarityComputer:
        return CosineSimilarityComputer(
            X, self.shrinkage, self.normalize, self.n_threads
        )


class AsymmetricCosineKNNConfig(BaseCosineKNNConfig):
    alpha: float = 0.5


class AsymmetricCosineKNNRecommender(BaseKNNRecommender):
    r"""K-nearest neighbor recommender system based on asymmetric cosine similarity. That is, the similarity matrix ``W`` is given by (column-wise top-k restricted)

    .. math::

        \mathrm{W}_{i,j} = \frac{\sum_{u} X_{ui} X_{uj}}{||X_{*i}||^{2\alpha}_2 ||X_{*j}||^{2(1-\alpha)}_2 + \mathrm{shrinkage}}

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        shrinkage (float, optional):
            The shrinkage parameter for regularization. Defaults to 0.0.
        alpha (bool, optional):
            Specifies :math:`\alpha`. Defaults to 0.5.
        top_k (int, optional):
            Specifies the maximal number of allowed neighbors. Defaults to 100.
        feature_weighting (str, optional):
            Specifies how to weight the feature. Must be one of:

                - "NONE" : no feature weighting
                - "TF_IDF" : TF-IDF weighting
                - "BM_25" : `Okapi BM-25 weighting <https://en.wikipedia.org/wiki/Okapi_BM25>`_

            Defaults to "NONE".
        bm25_k1 (float, optional):
            The k1 parameter for BM25. Ignored if ``feature_weighting`` is not "BM_25". Defaults to 1.2.
        bm25_b (float, optional):
            The b parameter for BM25. Ignored if ``feature_weighting`` is not "BM_25". Defaults to 0.75.
        n_threads (Optional[int], optional): Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if the variable is not set, it will be set to ``os.cpu_count()``. Defaults to None.

    """

    config_class = AsymmetricCosineKNNConfig

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        shrinkage: float = 0.0,
        alpha: float = 0.5,
        top_k: int = 100,
        feature_weighting: str = "NONE",
        bm25_k1: float = 1.2,
        bm25_b: float = 0.75,
        n_threads: Optional[int] = None,
    ):
        super().__init__(
            X_train_all,
            shrinkage,
            top_k,
            n_threads,
            feature_weighting=feature_weighting,
            bm25_k1=bm25_k1,
            bm25_b=bm25_b,
        )
        self.alpha = alpha

    def _create_computer(self, X: InteractionMatrix) -> AsymmetricSimilarityComputer:
        return AsymmetricSimilarityComputer(
            X, self.shrinkage, self.alpha, self.n_threads
        )


class JaccardKNNConfig(BaseKNNConfig):
    pass


class JaccardKNNRecommender(BaseKNNRecommender):
    r"""K-nearest neighbor recommender system based on Jaccard similarity. That is, the similarity matrix ``W`` is given by (column-wise top-k restricted)

    .. math::

        \mathrm{W}_{i,j} = \frac{ |U_i \cap U_j |}{ |U_i \cup U_j| + \mathrm{shrinkage}}

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        shrinkage (float, optional):
            The shrinkage parameter for regularization. Defaults to 0.0.
        top_k (int, optional):
            Specifies the maximal number of allowed neighbors. Defaults to 100.
        n_threads (Optional[int], optional): Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if the variable is not set, it will be set to ``os.cpu_count()``. Defaults to None.

    """

    config_class = JaccardKNNConfig

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        shrinkage: float = 0.0,
        top_k: int = 100,
        n_threads: Optional[int] = None,
    ) -> None:
        super().__init__(X_train_all, shrinkage, top_k, n_threads)

    def _create_computer(self, X: InteractionMatrix) -> JaccardSimilarityComputer:
        return JaccardSimilarityComputer(X, self.shrinkage, self.n_threads)


class TverskyIndexKNNConfig(BaseKNNConfig):
    alpha: float = 0.5
    beta: float = 0.5


class TverskyIndexKNNRecommender(BaseKNNRecommender):
    r"""K-nearest neighbor recommender system based on Tversky Index. That is, the similarity matrix ``W`` is given by (column-wise top-k restricted)

    .. math::

        \mathrm{W}_{i,j} = \frac{ |U_i \cap U_j |}{ |U_i \cap U_j | + \alpha |U_i \setminus U_j| + \beta |U_j \setminus U_i| + \mathrm{shrinkage}}

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        shrinkage (float, optional):
            The shrinkage parameter for regularization. Defaults to 0.0.
        alpha (float, optional):
            :math:`alpha` parameter. Defaults to 0.5.
        beta (float, optional):
            :math:`beta` parameter. Defaults to 0.5.
        top_k (int, optional):
            Specifies the maximal number of allowed neighbors. Defaults to 100.
        n_threads (Optional[int], optional): Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if the variable is not set, it will be set to ``os.cpu_count()``. Defaults to None.

    """

    config_class = TverskyIndexKNNConfig

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        shrinkage: float = 0.0,
        alpha: float = 0.5,
        beta: float = 0.5,
        top_k: int = 100,
        n_threads: Optional[int] = None,
    ) -> None:
        super().__init__(X_train_all, shrinkage, top_k, n_threads)
        self.alpha = alpha
        self.beta = beta

    def _create_computer(self, X: InteractionMatrix) -> TverskyIndexComputer:
        return TverskyIndexComputer(
            X, self.shrinkage, self.alpha, self.beta, self.n_threads
        )



from irspack.optimizers._optimizers import (
    AsymmetricCosineKNNOptimizer,
    AsymmetricCosineUserKNNOptimizer,
    BaseOptimizer,
    BaseOptimizerWithEarlyStopping,
    CosineKNNOptimizer,
    CosineUserKNNOptimizer,
    DenseSLIMOptimizer,
    EDLAEOptimizer,
    IALSOptimizer,
    JaccardKNNOptimizer,
    P3alphaOptimizer,
    RP3betaOptimizer,
    SLIMOptimizer,
    TopPopOptimizer,
    TverskyIndexKNNOptimizer,
)
from irspack.optimizers.autopilot import autopilot
from irspack.optimizers.base_optimizer import get_optimizer_class

__all__ = [
    "BaseOptimizer",
    "BaseOptimizerWithEarlyStopping",
    "TopPopOptimizer",
    "IALSOptimizer",
    "DenseSLIMOptimizer",
    "EDLAEOptimizer",
    "SLIMOptimizer",
    "P3alphaOptimizer",
    "RP3betaOptimizer",
    "CosineKNNOptimizer",
    "AsymmetricCosineKNNOptimizer",
    "CosineUserKNNOptimizer",
    "AsymmetricCosineUserKNNOptimizer",
    "JaccardKNNOptimizer",
    "TverskyIndexKNNOptimizer",
    "get_optimizer_class",
    "autopilot",
]
try:
    from ._optimizers import NMFOptimizer, TruncatedSVDOptimizer

    __all__.extend(["NMFOptimizer", "TruncatedSVDOptimizer"])
except ImportError:
    pass  # pragma: no cover

try:
    from ._optimizers import BPRFMOptimizer

    __all__.append("BPRFMOptimizer")
except ImportError:
    pass  # pragma: no cover

try:
    from ._optimizers import MultVAEOptimizer

    __all__.append("MultVAEOptimizer")
except ImportError:
    pass  # pragma: no cover
